﻿import time
import pymunk as pm
from pymunk import Vec2d
import pymunk.util as pymutil
import socket
import sys
import math


# warning DRAWPROPLINES not stable

HOST = sys.argv[1]
sendComandPort = int(sys.argv[2])
getDataPort =int(sys.argv[3])


COLID_sticky_ground = 1
COLID_ground = 2
COLID_goo = 3
COLID_verysticky_goo = 5
COLID_noactiv_goo = 4
COLID_Finish_box = 6


max_length = 200    
min_length =  10   
rest_length=110  # TODO create animated lenght
stiffness =1000 #500
damping=20 #2



fps = 30.0
dt =  1.0/fps/2.0 

print sys.version
print pm.version


class Clock():
    def __init__(self):
        self.lasttime = time.time()
    def tick(self,fps):
        t = 1./fps
        ct = time.time()
        work = ct-self.lasttime
        sleep = t -  work
        #print t,work,sleep
        if sleep>0:
            time.sleep(sleep)
        self.lasttime  = time.time()
        
clock = Clock()

class PhyEngine(object):
    def __init__(self):
        self.space = pm.Space(iterations=100)
        self.space.gravity = (0.0, -800.0)
        self.sphgoo = {}
        self.lines = {}
        
        self.pollcomand = ''
        self.RUNNING = True
        
        self.soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.comandsoc = socket.socket(socket.AF_INET,  socket.SOCK_DGRAM)
        self.comandsoc.bind((HOST, sendComandPort))
        self.comandsoc.setblocking(1)
        self.comandsoc.settimeout(0.001)
        
        self.finishbox = None
        self.gooInsideFinishbox = False
        
        def post_solve(space,arb):
            b =  arb._get_shapes()[0]
            if arb.is_first_contact:
                self.pollcomand += 'soundsticky|'
            bpos = b.body.position
            for c in arb.contacts:
                x =   c.position-bpos
                b.body.apply_impulse(x*7) 
                b.body.angular_velocity = 0
        def post_solveStop(space,arb):
            b =  arb._get_shapes()[0] 
            b.body.velocity = (0,0)    
            b.body.angular_velocity = 0

            
        self.space.add_collision_handler(COLID_goo, COLID_sticky_ground, None, None,post_solve,  None) 
        self.space.add_collision_handler(COLID_verysticky_goo, COLID_sticky_ground, None, None,post_solveStop,  None) 
        
        
        def pre_solve(*e):
            return False
        self.space.add_collision_handler(COLID_noactiv_goo, COLID_noactiv_goo, None, pre_solve, None, None)     
        self.space.add_collision_handler(COLID_noactiv_goo, COLID_goo, None, pre_solve, None, None) 
        
        
        def pre_solve_f(sep,arb):
            goo,p =  arb._get_shapes()

            for c in arb.contacts:
                x =   c.position-goo.body.position
                goo.body.apply_impulse(x*10) 
            
            
            
            return False
        
 
        
        self.space.add_collision_handler(COLID_goo, COLID_Finish_box, None, pre_solve_f, None, None) 
        self.space.add_collision_handler(COLID_verysticky_goo, COLID_Finish_box, None, pre_solve_f, None, None) 
        
        
    def addStaticPoly(self,points,collision_type = COLID_sticky_ground)  :
        points = map(Vec2d,points) 
        points = pymutil.convex_hull(points)
        if not pymutil.is_clockwise(points):
            points.reverse()
        #points = pymutil.poly_vectors_around_center(points)
        print 'stat add:'
        print    points
        body = pm.Body(pm.inf, pm.inf)
        body.position =  Vec2d(0,0)
        shape = pm.Poly(body, points, Vec2d(0,0))
        shape.collision_type = collision_type
        shape.friction =  1
        shape.elasticity = 0.9
        self.space.add_static(shape)
        return shape
        
    def addSphGoo(self,index,pos,radius = 11,mass = 1.0,collision_type = COLID_goo):
        print 'add goo pos:',pos
        moment = pm.moment_for_circle(mass, radius, 0.0, Vec2d(0,0))
        ball_body = pm.Body(mass, moment)
        ball_body.position = Vec2d(pos)
                
        ball_shape = pm.Circle(ball_body, radius, Vec2d(0,0))
        ball_shape.friction =  1
        ball_shape.elasticity = 0.9
        ball_shape.collision_type = collision_type
        self.space.add(ball_body, ball_shape)        
        self.sphgoo[index] = ball_shape
        print 'addSphGoo',index,ball_shape
        
    def toModel(self,indexgoo,indexgoo1,indexgoo2):
        b = self.sphgoo.get(indexgoo,None)
        b1 = self.sphgoo[indexgoo1]
        b2 = self.sphgoo[indexgoo2]

        z = 5  # b.radius/2
        
        #if not b or ( (b1,b2) not in  self.lines and (b2,b1) not in  self.lines ) :
        if not b :
            ds = pm.DampedSpring(b1.body, b2.body,(0,z), (0,-z),rest_length, stiffness, damping)
            self.space.add(ds)
            self.lines[(b2,b1)] = ds
            self.pollcomand += 'addbeam*%s*%s|'%(indexgoo1,indexgoo2)

        else:
            ds1 = pm.DampedSpring(b.body, b1.body,(0,z), (0,-z),rest_length, stiffness, damping)
            self.space.add(ds1)
            self.lines[(b,b1)] = ds1
            self.pollcomand += 'addbeam*%s*%s|'%(indexgoo,indexgoo1)
            
            ds2 = pm.DampedSpring(b.body, b2.body,(0,z), (0,-z),rest_length, stiffness, damping)
            self.space.add(ds2)
            self.lines[(b,b2)] = ds2
            self.pollcomand += 'addbeam*%s*%s|'%(indexgoo,indexgoo2)
 
    def queryforPoly(self,a,b):
        u"""if poly in segment a-b return True"""
        x = self.space.segment_query(a, b)
        if x:
            for qinfo in x:
                shape = qinfo.shape
                if type(shape) == pm.Poly and shape.collision_type != COLID_Finish_box:
                    return True
        

    def remove(self,index):
        goo = self.sphgoo[index]
        torem = []
        for key,ds in  self.lines.iteritems():
            if goo in key:
                self.space.remove(ds)
                torem.append(key)
                
        for key in torem:
            del  self.lines[key]
        del self.sphgoo[index]    
        self.space.remove(goo,goo.body)
        
    def removebyshapes(self,s):
        for index,shape in  self.sphgoo.iteritems():
            if s==shape:
                if shape.collision_type == COLID_verysticky_goo:
                    self.pollcomand += 'thisnodelete*%s|'%index
                    if s is not None and type(shape) not in( pm.Segment,pm.Poly):
                        s.body.apply_impulse((0,550),(-20,10))
                else:
                    self.remove(index)
                    self.pollcomand += 'delete*%s|plusgoo*|'%index
                return
              
               
                
    def removeAll(self):
        for x in self.space.constraints:
            self.space.remove(x)
            
        for x in self.space.bodies:
            self.space.remove(x)
        for x in self.space.shapes:
            self.space.remove(x)
        for x in self.space.static_shapes:
            self.space.remove_static(x)
        self.sphgoo = {}
        self.lines = {}
        self.finishbox = None
        self.gooInsideFinishbox = False
        
            
    def getDikCoordGoo(self):
        d = {}
        #print self.sphgoo
        for index,ch in  self.sphgoo.iteritems():
            #print 'index,ch',index,ch
            pos = list(ch.body.position)
            angle = ch.body.angle *180/math.pi
            d[index] = (pos,angle)
        return d

    def commandEval(self,datastr):
       # print datastr
        for x in datastr.split('|'):
            if not x:
                continue
            elif x.startswith('close'):
                self.RUNNING = False        
            elif x.startswith('info*'):
                t = time.time()- float(x.split('*')[1])
                print '### info ###'
                print 'time ',t
                print 'constraints:',len(self.space.constraints)
                print 'bodies:',len(self.space.bodies)
                print 'shapes:',len(self.space.shapes)
                print 'static_shapes:',len(self.space.static_shapes)
                print 'static_body:',self.space.static_body  
                print '### ###'
                
            elif x.startswith('addstatic*'):
                points = eval(x.split('*')[1])
                self.addStaticPoly(points)
            elif x.startswith('addfinishbox*'):
                points = eval(x.split('*')[1])
                self.finishbox = self.addStaticPoly(points,collision_type=COLID_Finish_box)    
                
            elif x.startswith('addgoo*'):
                data = eval(x.split('*')[1]) # data = {'index':int,'pos':[x,y],'radius' = 11,'mass' = 1.0 }
                self.addSphGoo(**data)
                
            elif x.startswith('toModel*'):
                data =  eval(x.split('*')[1]) #data = [index,index1,index2]: all int
                self.toModel( *data)
            elif x.startswith('jumpbycoord*'):
                ccor = map(float,x.split('*')[1:])
                shape = self.space.point_query_first(Vec2d(ccor))
                if shape is not None and type(shape) not in( pm.Segment,pm.Poly):
                    shape.body.apply_impulse((800,500),(50,10))           
            elif x.startswith('deletebycoord*'):
                ccor = map(float,x.split('*')[1:])
                shape = self.space.point_query_first(Vec2d(ccor))
                if shape:
                    self.removebyshapes(shape)
                

                    
            elif x.startswith('movegoo*'):
                ccor = map(float,x.split('*')[1:])            
                ok,a,b = self.markGooPosible(ccor)
                if not b:
                    continue
                elif not ok:
                    self.pollcomand += 'drawline*%s|'%str([list(a[1].body.position),  list(b[1].body.position)])
                elif ok:
                    self.pollcomand += 'drawline*%s|'%str([ccor,list(b[1].body.position)])
                    self.pollcomand += 'drawline*%s|'%str([ccor,list(a[1].body.position)])
            elif x.startswith('gooconn*'):   #
                index = int(x.split('*')[1])
                ccor = map(float,x.split('*')[2:])            
                ok,a,b = self.markGooPosible(ccor)
                if not b:
                    #print 'HRENATENYSKI !!!!!'
                    self.pollcomand += 'stelsdelete*%s|'%index
                elif not ok:
                    self.toModel(0,a[0],b[0])
                    #self.pollcomand += 'minusgoo*|stelsdelete*%s|soundcreate|'%index
                    # ok, no del goo
                    self.pollcomand += 'stelsdelete*%s|soundcreate|'%index
                elif ok:
                    self.addSphGoo(index,ccor)
                    self.toModel(index,a[0],b[0])
                    self.pollcomand += 'minusgoo*|'
            elif x.startswith('gooconnforce*'):
                a,b = map(int,x.split('*')[1:])
                self.toModel(0,a,b)    
            elif x.startswith('removeall'):
                self.removeAll()
                
                
            
            else:
                print 'phy_engine:unknown command'
                print x
                
    def markGooPosible(self,coord):
        coord = Vec2d(coord)
        if len(self.sphgoo)<2:
            return 0,0,0
        ls = self.sphgoo.items()
        ls.sort(key= lambda a: (a[1].body.position-coord).get_length())
        index1,b1 = ls[0]
        index2,b2 = ls[1]
        
        l1 = (b1.body.position-coord).get_length()
        l2 = (b2.body.position-coord).get_length()
        if l1<min_length or l2>max_length:
            return 0,0,0

        
        if ( (b1,b2) not in  self.lines and (b2,b1) not in  self.lines ) :
            if self.queryforPoly(b1.body.position,b2.body.position):
                return 0,0,0
            return 0,[index1,b1],[index2,b2]
        else:
            if self.queryforPoly(coord,b1.body.position) or \
                self.queryforPoly(coord,b2.body.position):
                return 0,0,0
            return 1,[index1,b1],[index2,b2]
        
        
    def run(self):
        while self.RUNNING:
            self.space.step(dt)
            self.space.step(dt)
            
            if self.finishbox:
                x = self.space.shape_query( self.finishbox)
                if self.gooInsideFinishbox and not x:
                    self.gooInsideFinishbox = False
                    self.pollcomand+= 'goooutbox*|'
                if not self.gooInsideFinishbox and  x:    
                    self.gooInsideFinishbox = True
                    self.pollcomand+= 'gooinbox*|'
                
            
            
            dik = {}
            {'coord':{},'lines':{},'comand':'','time':0}
            dik['coord'] = self.getDikCoordGoo()
            dik['time']  = time.time()
            dik['comand'] = self.pollcomand
            self.pollcomand = ''
            self.soc.sendto(str(dik),(HOST,getDataPort ))
        
            
            try:
                data, address = self.comandsoc.recvfrom(4096) 
                if address[0] != HOST:
                    print 'PHY:-Warning, unknown address(ignor comand):',address    
                else:
                    self.commandEval(data)
            except socket.timeout:
                pass  


 
            clock.tick(fps)    
                    
        self.soc.close()
        self.comandsoc.close()
        print 'by by'
        sys.exit(1)
    

 

        
PhyEngine().run()       
        